from .int_to_int import int_to_int

def ienumerate(iterator, start=0, key=int_to_int):
    """
    meningkatkan fungsi enumerate() pada python
    untuk key menggunakan huruf dan basis angka lainnya.

    ```python
    it = ["ini", "contoh", "enumerator"]
    print(ienumerate(it))
    iprint(ienumerate(it, key=int_to_chr))
    ```
    """
    for i, v in enumerate(iterator, start):
        yield (key(i), v)



from .filter_empty import filter_empty
from .is_iterable import is_iterable
from .to_str import to_str


def ijoin(
    iterable,
    separator="",
    start="",
    end="",
    remove_empty=False,
    recursive=True,
    recursive_flat=False,
    str_strip=False,
):
    """
    Simplify Python join functions like PHP function.
    Iterable bisa berupa sets, tuple, list, dictionary.

    ```python
    arr = {'asd','dfs','weq','qweqw'}
    print(ijoin(arr, ', '))

    arr = '/ini/path/seperti/url/'.split('/')
    print(ijoin(arr, ','))
    print(ijoin(arr, ',', remove_empty=True))

    arr = {'a':'satu', 'b':(12, 34, 56), 'c':'tiga', 'd':'empat'}
    print(ijoin(arr, separator='</li>\\n<li>', start='<li>', end='</li>',
        recursive_flat=True))
    print(ijoin(arr, separator='</div>\\n<div>', start='<div>', end='</div>'))
    print(ijoin(10, ' '))
    ```
    """
    if not is_iterable(iterable):
        iterable = [iterable]

    separator = to_str(separator)

    if isinstance(iterable, dict):
        iterable = iterable.values()

    if remove_empty:
        # iterable = (i for i in filter_empty(iterable))
        iterable = filter_empty(iterable)

    if recursive:
        rec_flat = dict(start=start, end=end)
        if recursive_flat:
            rec_flat = dict(start="", end="")

        def rec(x):
            return ijoin(
                iterable=x,
                separator=separator,
                **rec_flat,
                remove_empty=remove_empty,
                recursive=recursive,
                recursive_flat=recursive_flat,
            )

        iterable = ((rec(i) if is_iterable(i) else i) for i in iterable)

    # iterable = (str(i) for i in iterable)
    iterable = map(str, iterable)

    if str_strip:
        # iterable = (i.strip() for i in iterable)
        iterable = map(str.strip, iterable)

    result = start

    for index, value in enumerate(iterable):
        if index:
            result += separator
        result += value

    result += end

    return result



import sqlite3


def sqlite_get_all_tables(filename):
    """
    Perintah SQLite untuk menampilkan seluruh tabel
    yang ada pada database.
    Hanya akan mengembalikan kolom nama tabel saja.
    """
    con = sqlite3.connect(filename)
    cur = con.cursor()
    cur.execute("SELECT name FROM sqlite_master WHERE type='table' order by name;")
    tables = cur.fetchall()
    con.close()
    return tables



import traceback


def traceback_filename(stack_level=-3):
    """
    Mendapatkan filename dimana fungsi yg memanggil
    fungsi dimana fungsi ini diletakkan dipanggil.

    ```py
    print(traceback_filename())
    ```
    """
    return traceback.extract_stack()[stack_level].filename



import os
import shutil


def django_clear_migrations(appname):
    migrations_path = os.path.join(appname, "migrations")
    if os.path.exists(migrations_path):
        shutil.rmtree(migrations_path)
    os.makedirs(migrations_path)
    open(os.path.join(migrations_path, "__init__.py"), "w").close()



from pypipr import *


def input_parameter(o):
    try:
        return eval(o)
    except Exception:
        return o



def main(module="pypipr"):
    module = __import__(module)
    m = ivars(module)
    m = m["module"] | m["variable"] | m["class"] | m["function"]
    # m = m["variable"] | m["function"]
    m = [x for x in m]
    m.sort()

    a = iargv(1)
    p = "Masukan Nomor Urut atau Nama Fungsi : "
    m = choices(daftar=m, contains=a, prompt=p)

    f = getattr(module, m)

    if a != m:
        print_colorize(m)
        print(f.__doc__)


    if inspect.isclass(f):
        print_log("Class tidak dapat dijalankan.")
    elif inspect.ismodule(f):
        print_log("Module tidak dapat dijalankan.")
        main(f.__name__)
        return
    elif inspect.isfunction(f):
        s = inspect.signature(f)

        if not a:
            print(m, end="")
            print_colorize(s)

        k = {}
        for i, v in s.parameters.items():
            o = iinput(f"{i} [{v.default}] : ")
            if len(o):
                try:
                    k[i] = input_parameter(o)
                except Exception:
                    print_colorize(
                        "Input harus dalam syntax python.",
                        color=colorama.Fore.RED,
                    )

        f = f(**k)

    else:
        # variable
        pass

    iprint(f)


if __name__ == "__main__":
    main()



from .is_empty import is_empty


def to_str(value):
    """
    Mengubah value menjadi string literal

    ```python
    print(to_str(5))
    print(to_str([]))
    print(to_str(False))
    print(to_str(True))
    print(to_str(None))
    ```
    """
    if isinstance(value, str):
        return value
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, bool):
        return "1" if value else "0"
    if is_empty(value):
        return ""
    try:
        return str(value)
    except Exception:
        raise Exception(f"Tipe data {value} tidak diketahui")



def avg(i):
    """
    Simple Average Function karena tidak disediakan oleh python

    ```python
    n = [1, 22, 2, 3, 13, 2, 123, 12, 31, 2, 2, 12, 2, 1]
    print(avg(n))
    ```
    """
    return sum(i) / len(i)



import json

import yaml

from .idumps_html import idumps_html


def idumps(data, syntax="yaml", indent=4):
    """
    Mengubah variabel data menjadi string untuk yang dapat dibaca untuk disimpan.
    String yang dihasilkan berbentuk syntax YAML/JSON/HTML.

    ```python
    data = {
        'a': 123,
        't': ['disini', 'senang', 'disana', 'senang'],
        'l': (12, 23, [12, 42]),
    }
    print(idumps(data))
    print(idumps(data, syntax='html'))
    ```
    """
    if syntax == "yaml":
        return yaml.dump(data, indent=indent)
    if syntax == "json":
        return json.dumps(data, indent=indent)
    if syntax == "html":
        return idumps_html(data, indent=indent)
    raise Exception(f"Syntax tidak didukung {syntax}")



from .ComparePerformance import ComparePerformance
from .LINUX import LINUX
from .PintUreg import PintUreg
from .PintUregQuantity import PintUregQuantity
from .RunParallel import RunParallel
from .WINDOWS import WINDOWS
from .auto_reload import auto_reload
from .avg import avg
from .basename import basename
from .batch_calculate import batch_calculate
from .batchmaker import batchmaker
from .bin_to_int import bin_to_int
from .calculate import calculate
from .choices import choices
from .chr_to_int import chr_to_int
from .chunk_array import chunk_array
from .console_run import console_run
from .create_folder import create_folder
from .datetime_from_string import datetime_from_string
from .datetime_now import datetime_now
from .dict_first import dict_first
from .dirname import dirname
from .django_clear_migrations import django_clear_migrations
from .django_runserver import django_runserver
from .exit_if_empty import exit_if_empty
from .filter_empty import filter_empty
from .get_by_index import get_by_index
from .get_class_method import get_class_method
from .get_filemtime import get_filemtime
from .get_filesize import get_filesize
from .github_init import github_init
from .github_pull import github_pull
from .github_push import github_push
from .github_user import github_user
from .hex_to_int import hex_to_int
from .iargv import iargv
from .idir import idir
from .idumps import idumps
from .idumps_html import idumps_html
from .ienumerate import ienumerate
from .ienv import ienv
from .iexec import iexec
from .iinput import iinput
from .ijoin import ijoin
from .iloads import iloads
from .iloads_html import iloads_html
from .input_char import input_char
from .int_to_bin import int_to_bin
from .int_to_chr import int_to_chr
from .int_to_hex import int_to_hex
from .int_to_int import int_to_int
from .int_to_oct import int_to_oct
from .iopen import iopen
from .iprint import iprint
from .irange import irange
from .ireplace import ireplace
from .is_empty import is_empty
from .is_html import is_html
from .is_iterable import is_iterable
from .is_raw_string import is_raw_string
from .is_valid_url import is_valid_url
from .iscandir import iscandir
from .isplit import isplit
from .ivars import ivars
from .log import log
from .no_indent import no_indent
from .oct_to_int import oct_to_int
from .password_generator import password_generator
from .path_to_module import path_to_module
from .pip_freeze_without_version import pip_freeze_without_version
from .pip_update_pypipr import pip_update_pypipr
from .poetry_publish import poetry_publish
from .poetry_update_version import poetry_update_version
from .print_colorize import print_colorize
from .print_dir import print_dir
from .print_log import print_log
from .print_to_last_line import print_to_last_line
from .random_bool import random_bool
from .repath import repath
from .restart import restart
from .set_timeout import set_timeout
from .sets_ordered import sets_ordered
from .sqlite_backup import sqlite_backup
from .sqlite_delete_table import sqlite_delete_table
from .sqlite_get_all_tables import sqlite_get_all_tables
from .sqlite_get_data_table import sqlite_get_data_table
from .str_cmp import str_cmp
from .text_colorize import text_colorize
from .tiles import tiles
from .to_str import to_str
from .traceback_filename import traceback_filename
from .traceback_framename import traceback_framename
import asyncio
import colorama
import csv
import datetime
import functools
import inspect
import io
import json
import lxml
import math
import multiprocessing
import operator
import os
import pathlib
import pint
import pprint
import queue
import random
import re
import requests
import string
import subprocess
import sys
import textwrap
import threading
import time
import traceback
import tzdata
import uuid
import webbrowser
import yaml
import zoneinfo


from queue import Queue

from .console_run import console_run
from .RunParallel import RunParallel


def collectstatic():
    console_run("python manage.py collectstatic --noinput")


def makemigrations():
    console_run("python manage.py makemigrations")


def migrate():
    console_run("python manage.py migrate")


def runserver():
    console_run("python -Wa manage.py runserver 0.0.0.0:8080")


class RP(RunParallel):
    def static_update(self, result: dict, q: Queue):
        collectstatic()

    def database_update(self, result: dict, q: Queue):
        makemigrations()
        migrate()


def django_runserver():
    try:
        RP().run_multi_threading()
        runserver()
    except KeyboardInterrupt:
        pass



import os

from pypipr.dirname import dirname


def path_to_module(path, indeks=0):
    """
    Mengubah absolute path file menjadi path modul relatif terhadap cwd (current working directory),
    dengan opsi untuk memangkas bagian akhir path berdasarkan indeks.

    Parameter:
        abs_path (str): Path absolut menuju file.
        indeks (int):
            - 0 => hasil lengkap hingga file (tanpa ekstensi),
            - -1 => tanpa nama file, hanya foldernya,
            - -2 => dua folder di atasnya, dst.

    Returns:
        str: Path bergaya modul Python (dipisah dengan ".")
    """
    path = dirname(path, abs_path=False, indeks=indeks)
    return path.replace(os.sep, ".")

    cwd = os.getcwd()
    rel_path = os.path.relpath(abs_path, cwd)
    rel_path_no_ext = os.path.splitext(rel_path)[0]

    parts = rel_path_no_ext.split(os.sep)

    if indeks < 0:
        parts = parts[: len(parts) + indeks]

    return ".".join(parts)



import traceback


def traceback_framename(stack_level=-3):
    """
    Mendapatkan frame name dimana fungsi yg memanggil
    fungsi dimana fungsi ini diletakan ini dipanggil.

    ```py
    print(traceback_framename())
    ```
    """
    return traceback.extract_stack()[stack_level].name



from .console_run import console_run


def poetry_publish(token=None):
    """
    Publish project to pypi,org

    ```py
    poetry_publish()
    ```
    """
    if token:
        console_run("update token", f"poetry config pypi-token.pypi {token}")
    console_run("Build", "poetry build")
    console_run("Publish to PyPi.org", "poetry publish")



import lxml

from .ijoin import ijoin


def iloads_html(html):
    """
    Mengambil data yang berupa list `<ul>`, dan table `<table>` dari html
    dan menjadikannya data python berupa list.
    setiap data yang ditemukan akan dibungkus dengan tuple sebagai separator.

    ```
    list (<ul>)     -> list         -> list satu dimensi
    table (<table>) -> list[list]   -> list satu dimensi didalam list
    ```
    
    apabila data berupa ul maka dapat dicek type(data) -> html_ul
    apabila data berupa ol maka dapat dicek type(data) -> html_ol
    apabila data berupa dl maka dapat dicek type(data) -> html_dl
    apabila data berupa table maka dapat dicek type(data) -> html_table

    ```python
    import pprint
    pprint.pprint(iloads_html(iopen("https://harga-emas.org/")), depth=10)
    pprint.pprint(iloads_html(iopen("https://harga-emas.org/1-gram/")), depth=10)
    ```
    """

    def xpath(e, x):
        """
        Fungsi ini sangat krusial/menentukan. Fungsi ini dibuat
        supaya xpath yang diberikan diproses dari element saat ini.
        Sedangkan xpath pada element lxml akan mengecek syntax xpath dari
        root paling awal document.
        """
        if not isinstance(e, str):
            e = lxml.html.tostring(e, encoding="unicode")
        e = lxml.html.fromstring(e)
        return (e, e.xpath(x))

    def parse(e):
        parser = {
            "ul": parse_ul,
            "ol": parse_ol,
            "dl": parse_dl,
            "table": parse_table,
        }
        try:
            return parser[e.tag.lower()](e)
        except Exception:
            raise Exception("Tidak ditemukan parse fungsi untuk element : ", e)

    def parse_list(ul):
        """
        Simple parse html list.
        """
        result = []
        _, li = xpath(ul, "li")
        for i in li:
            u = iloads_html(i)
            t = i.text.strip() if i.text else ""
            if t and t != u:
                result.append({t: u})
            else:
                result.append(u)
        return result

    def parse_ul(ul):
        return html_ul(parse_list(ul))

    def parse_ol(ol):
        return html_ol(parse_list(ol))

    def parse_dl(dl):
        """
        Simple parse dl-dt-dd.
        """
        result = html_dl()
        _, di = xpath(dl, "dt|dd")
        d = iter(di)
        try:
            while True:
                i = next(d)
                k = i.tag.lower()
                v = iloads_html(i)
                if k == "dt":
                    result.append(["", []])
                    result[-1][0] = v
                elif k == "dd":
                    result[-1][-1].append(v)
        except Exception:
            pass

        return result

    def parse_table(table):
        """
        Mengambil children tr dari table.
        tr diambil dari thead atau tbody atau langsung tr.
        tr dari thead atau tbody akan dianggap sama.
        """
        result = html_table()
        _, tr = xpath(table, "//tr[not(ancestor::tr)]")
        for itr in tr:
            d = []
            _, td = xpath(itr, "th|td")
            for itd in td:
                d.append(iloads_html(itd))
            result.append(d.copy())
        return result

    def text_content(e):
        """
        mengambil semua text dalam element secara recursive.
        karena tidak ditemukan data dalam element.
        """
        return ijoin(e.itertext(), str_strip=True)

    element, childs = xpath(
        html,
        "//*[self::ul | self::ol | self::dl | self::table]"
        "[not(ancestor::ul | ancestor::ol | ancestor::dl | ancestor::table)]",
    )
    if childs:
        return tuple((parse(data) for data in childs))
    else:
        return text_content(element)


class html_ul(list):
    """
    Class ini digunakan untuk idumps dan iloads html
    """

    pass


class html_ol(list):
    """
    Class ini digunakan untuk idumps dan iloads html
    """

    pass


class html_dl(list):
    """
    Class ini digunakan untuk idumps dan iloads html
    """

    pass


class html_table(list):
    """
    Class ini digunakan untuk idumps dan iloads html
    """

    pass



def chr_to_int(s, start=0, numbers="abcdefghijklmnopqrstuvwxyz"):
    """
    Fungsi ini berguna untuk mengubah urutan huruf menjadi angka.

    ```python
    print(chr_to_int('z'))  # Output: 26
    print(chr_to_int('aa'))  # Output: 27
    print(chr_to_int('abc', numbers="abc"))  # Output: 18
    ```
    """
    result = 0
    digit = len(numbers)
    for char in s:
        result = result * digit + numbers.index(char) + 1
    return result + start - 1



import sys


def iargv(key: int, cast=None, on_error=None):
    """
    Mengambil parameter input dari terminal tanpa menimbulkan error
    apabila parameter kosong.
    Parameter yg berupa string juga dapat diubah menggunakan cast.

    ```python
    print(iargv(1, cast=int, on_error=100))
    ```
    """
    try:
        v = sys.argv[key]
        if cast:
            return cast(v)
        else:
            return v
    except Exception:
        return on_error



def int_to_bin(n):
    """
    Fungsi ini sama seperti fungsi bin().
    fungsi ini dibuat hanya untuk keperluan pembuatan module semata.

    ```python
    print(int_to_bin(7777))
    ```
    """
    return bin(n)



import sys

argv = sys.argv[1:]
argv_iter = iter(argv)


def iinput(*args, **kwargs):
    """
    iinput()
    iinput("pesan")
    iinput("pesan", default)
    iinput(msg="...", default="...")
    iinput(default="...")
    """
    global argv_iter
    k = {}

    try:
        _next = next(argv_iter)
        if _next != "_":
            return _next
        k["_"] = True
    except Exception:
        pass


    def get_args(indeks):
        try:
            return args[indeks]
        except Exception:
            raise

    def get_kwargs(key, indeks):
        if key in kwargs:
            return kwargs[key]
        return get_args(indeks)

    try:
        k["msg"] = get_kwargs("msg", 0)
    except Exception:
        pass

    try:
        k["default"] = get_kwargs("default", 1)
    except Exception:
        pass

    use_default = "_" in k
    no_msg = "msg" not in k
    no_default = "default" not in k
    has_msg = "msg" in k
    has_default = "default" in k

    if use_default and has_default:
        return k["default"]

    if has_default and no_msg:
        return k["default"]

    if has_msg and no_default:
        return input(k["msg"])

    if has_msg and has_default:
        return input(k["msg"]) or k["default"]



from math import ceil


def tiles(
    tile_x=None,
    tile_y=None,
    area_x=None,
    area_y=None,
    gap_x=None,
    gap_y=None,
):
    x = ceil(area_x / (tile_x + gap_x))
    y = ceil(area_y / (tile_y + gap_y))
    return x * y


if __name__ == "__main__":
    print(ceil(8/8))
    print(tiles(30, 30, 300, 300, 10, 10))



def print_to_last_line(text: str, latest=1, clear=True):
    """
    Melakukan print ke konsol tetapi akan menimpa baris terakhir.
    Berguna untuk memberikan progress secara interaktif.

    ```python
    for i in range(5):
        print(str(i) * 10)
    print_to_last_line(f" === last ===")
    ```
    """
    t = ""

    if latest:
        t += f"\033[{latest}A"

    # if append:
    #     t += "\033[10000C" 
    # else:
    #     # t += "\r"
    #     t += "\033[F"
    #     # if clear:
    #     #     t += "\033[K"

    if clear:
        t += "\033[K"

    t += text
    print(t)


if __name__ == "__main__":
    c = input("masukkan apa saja : ")
    print_to_last_line(f"masukkan apa saja : {c} [ok]")
    print(f"masukkan apa saja : ")
    print(f"masukkan apa saja : bsmsidoneoslwjsj ")
    print(f"masukkan apa saja : ")
    print_to_last_line(f"masukkan apa saja : {c} [ok]", latest=2, clear=False)
